# Group is a facility for assigning the same permssions to several users.
#
# Note that if 'group1' and 'group2' are configured identically, then users
# in the two groups will have identical permissions.  There is nothing
# special about being 'in a group'; it is merely a convenient way to 
# assign identical permissions.
class Group < ActiveRecord::Base
  concerned_with :authentication

  LOCAL_DATABASE = "Local Database"  
  
  has_many :group_users, :dependent => :destroy, :include => :user
  has_many :paths, :dependent => :destroy, :class_name => "GroupPath", :include => [:end_point_1, :end_point_2]

  has_many :users, :through => :group_users do
    # :call-seq: 
    #   group.users.unattached
    #
    # Returns all non-admin Users NOT attached to this Group
    def unattached
      (User.non_admin - self).sort_by(&:name_and_login)
    end
  end
  
  has_many :logins, :through => :group_logins do
    # :call-seq:
    #   group.logins.unattached
    #
    # Returns (node) logins that are NOT attached to this Group.
    def unattached
      Login.node - self
    end
  end
  
  validates_presence_of :name, :directory
  validates_uniqueness_of :name
  validates_uniqueness_of :urn, :allow_blank => true
  validates_length_of :name, :within => 1..255, :allow_blank => true
  validates_length_of :description, :maximum => 5000

  # Returns +true+ if _node_ in included by any of the paths.
  def include_node?(node)
    paths.any? { |path| path.include_node?(node) }
  end

  # if urn is present, this signifies that the user has come from remote authentication
  def remote?
    urn?
  end

  ## Failover Group Homogenization ###

  # Select all paths that are cluster equivalent to given path
  def get_cluster_equivalent_node_group_paths(path)
    all_paths = []

    paths.each do |p|
      if path.end_point_1.cluster_equivalent?(p.end_point_1)
        if path.end_point_2.cluster_equivalent?(p.end_point_2)
          all_paths << p
        end

      # Check reverse endpoints if bidirectional
      elsif path.direction == GroupPath::DIRECTION_TO_AND_FROM && p.direction == GroupPath::DIRECTION_TO_AND_FROM
        if path.end_point_1.cluster_equivalent?(p.end_point_2)
          if path.end_point_2.cluster_equivalent?(p.end_point_1)
            all_paths << p
          end
        end
      end
    end

    all_paths
  end

  def generate_node_group_paths(path)
    return nil if path.end_point_1.nil? || path.end_point_2.nil?

    ep1 = path.end_point_1
    ep2 = path.end_point_2

    # Any -> Any
    if ep1.star_at_star? && ep2.star_at_star?
      return [path]
    end

    all_paths = []

    if !ep1.star_at_star? && !ep2.star_at_star? && ep1.owner == ep2.owner
      all_paths << path
    end

    # Both endpoints in same group
    # Match ep1 -> [ep2, ep3...]
    if path.endpoints_in_same_node_group?
      other_eps = ep1.cluster_equivalent_endpoints

      other_eps.each do |other_ep|
        new_path = path.clone_path(ep1, other_ep)

        # Merge redundant unidirectional relationships
        if path.direction == GroupPath::DIRECTION_TO || path.direction == GroupPath::DIRECTION_FROM
          new_path.direction = GroupPath::DIRECTION_TO_AND_FROM
        end

        all_paths << new_path
      end
    else
      # Endpoints in different groups
      # Match each endpoint1 to each endpoint2
      eps1 = ep1.cluster_equivalent_endpoints << ep1
      eps2 = ep2.cluster_equivalent_endpoints << ep2

      eps1.each do |e1|
        eps2.each do |e2|
          if e1.address != e2.address
            all_paths << path.clone_path(e1, e2)
          end
        end
      end
    end

    all_paths
  end

  # Find an existing endpoint on the node that corresponds to the given endpoint
  def get_corresponding_endpoint(node, other_endpoint)
    node.endpoints.each do |ep|
      return ep if ep.login == other_endpoint.login
    end
  end

  # Check if a the given path already exists in the group
  def path_exists?(path)
    paths.any? { |other_path| path.paths_equivalent?(other_path) }
  end

  # Find paths containing other_node's endpoints and duplicate with corresponding node endpoint
  def homogenize_node_group_paths(node, other_node)
    return if node.nil? || other_node.nil?

    new_paths = []

    paths.each do |p|
      p_ep1 = p.end_point_1
      p_ep2 = p.end_point_2

      next if p_ep1.nil? || p_ep2.nil?

      # Note: Can't call owner if Any (star_at_star?), must check

      if !p_ep1.star_at_star? && p_ep1.owner == other_node
        if p_ep2.star_at_star? || p_ep2.owner != node
          # First endpoint match
          new_endpoint = get_corresponding_endpoint(node, p.end_point_1)

          new_path = p.clone_path(new_endpoint, p_ep2)

          new_paths << new_path unless path_exists?(new_path)
        end
      elsif !p_ep2.star_at_star? && p_ep2.owner == other_node
        if p_ep1.star_at_star? || p_ep1.owner != node
          # Second endpoint match
          new_endpoint = get_corresponding_endpoint(node, p.end_point_2)
          new_path = p.clone_path(p_ep1, new_endpoint)

          new_paths << new_path unless path_exists?(new_path)
        end
      end
    end

    self.paths += new_paths unless new_paths.nil?
  end

  ######

  class << self
    
    # Creates default groups / paths.  This is presumably called at installation
    # time to create groups with admin, transfer, and monitor rights on all nodes.
    def create_installation_default_groups
      create_admin_all_nodes_group
      create_initiator_all_nodes_group
      create_monitor_all_nodes_group
    end
    
    def create_admin_all_nodes_group
        group_name = " ALL NODES - Transfer Administrators"
        return if find_by_name(group_name)
        group = Group.create(
          :name => group_name,
          :description => "Sees and controls all transfers on all nodes",
          :directory => LOCAL_DATABASE
          )
        group.paths.build(
          :end_point_1        => Login.star_at_star,
          :direction          => GroupPath::DIRECTION_TO_AND_FROM,
          :end_point_2        => Login.star_at_star,
          :start_smart        => true,
          :start_simple       => true,
          :create_smart       => true,
          :share_smart        => true,
          :control            => true,
          :view               => true,
          :email_notification => true
        )
        group.save
    end

    def create_initiator_all_nodes_group
      group_name = " ALL NODES - Transfer Initiators"
      return if find_by_name(group_name)
      group = Group.create(
        :name => group_name,
        :description => "Can start transfers on all nodes, sees and controls transfers started by self",
        :directory => LOCAL_DATABASE
        )
      group.paths.build(
        :end_point_1        => Login.star_at_star,
        :direction          => GroupPath::DIRECTION_TO_AND_FROM,
        :end_point_2        => Login.star_at_star,
        :start_smart        => true,
        :start_simple       => true,
        :create_smart       => true,
        :share_smart        => false,
        :control            => false,
        :view               => false,
        :email_notification => false
      )
      group.save!
    end
    
    def create_monitor_all_nodes_group
      group_name = " ALL NODES - Transfer Monitors"
      return if find_by_name(group_name)
      group = Group.create(
        :name => group_name,
        :description => "Sees all transfers on all nodes",
        :directory => LOCAL_DATABASE
        )
      group.paths.build(
        :end_point_1        => Login.star_at_star,
        :direction          => GroupPath::DIRECTION_TO_AND_FROM,
        :end_point_2        => Login.star_at_star,
        :start_smart        => false,
        :start_simple       => false,
        :create_smart       => false,
        :share_smart        => false,
        :control            => false,
        :view               => true,
        :email_notification => true
      )
      group.save
    end
    
    def find_remote(name)
      RemoteServices.get_groups(name)          
    end
      
  end

end
