class JobStep
  # Request a transfer token from the non initiating side
  def paths
    if shares_transfer?
      # AC - 677 - Destination always to be the file name and Destination root to be folder
      source_paths.multiline_split_and_compact.map { |path| { :source => path, :destination => File.basename(path) } }
    else
      source_paths.multiline_split_and_compact.map { |path| { :source => path } }
    end
  end

  def transfer_spec_request
    {
      :transfer_requests => [
        {
          :transfer_request => {
            :paths => paths,
             # AC - 677 - Destination always to be the file name and Destination root to be folder
             :destination_root => dest_path
          }
        }
      ]
    }
  end

  def transfer_request_options
    # Organized by section

    # Implicit
    options = {
      :authentication => :token,
      :cookie => cookie,
      :title => job.description,
    }

    # Connection
    options[:fasp_port] = udp_port if udp_port.present?
    options[:proxy] = proxy_url_with_credentials if use_fasp_proxy?
    options[:rate_policy] = policy_value.downcase if !policy_value.nil?
    # Transfer
    options.merge!({
      :min_rate_kbps => minimum_rate,
      :retry_count => max_retry_attempts,
      :retry_interval => retry_interval_seconds,
      :target_rate_kbps => target_rate
    })

    # Security
    if ear_unencrypted_password.present?
      options[:content_protection] = 'encrypt'
      options[:content_protection_password] = ear_unencrypted_password
    else
      options[:content_protection] = nil
    end
    options[:cipher] = encryption.downcase


    # File handling
    options[:exclude_newer_than] = -file_exclude_newer_than if file_exclude_by_timestamp

    if resume_partially_transferred_files?
      options[:resume_policy] = case file_compare_type
                                  when 'Attributes'
                                    'attrs'
                                  when 'Full Checksum'
                                    'full_csum'
                                  else
                                    'sparse_csum'
                                end
    else
      options[:resume_policy] = 'none'
    end

    options.merge!({
      :excludes => exclude_file_filter.to_s.split(',').map{|f| f.strip },
      :includes => include_file_filter.to_s.split(',').map{|f| f.strip },
      :overwrite => file_overwrite_policy.downcase,
      :preserve_access_time => preserve_file_timestamps? && preserve_file_access_times?,
      :preserve_creation_time => preserve_file_timestamps? && preserve_file_create_times?,
      :preserve_modification_time => preserve_file_timestamps? && preserve_file_modify_times?,
      :preserve_dates => preserve_file_timestamps? && preserve_file_source_access_times?,
      :preserve_user_id => preserve_file_uids?,
      :preserve_group_id => preserve_file_gids?,
      :symlink_policy => (symbolic_link_handling.presence || 'follow').downcase
    })

    options[:delete_before_transfer] = delete_before_transfer?

    unless use_ascp4?
      if archive_after_transfer?
        options[:move_after_transfer] = archive_directory
      else
        options[:remove_after_transfer] = delete_source_after_transfer? || delete_source_dirs_after_transfer?
        options[:remove_empty_source_directory] = delete_source_dirs_after_transfer?
      end
      options[:remove_empty_directories] = delete_empty_directories?
    end

    # Advanced
    options[:dgram_size] = datagram_size if datagram_size.present?
    options[:read_size] = read_block_size if read_block_size.present?
    options[:write_size] = write_block_size if write_block_size.present?
    options[:use_ascp4] = use_ascp4?
    # AC-846 FIX - pass Specify base for source path
    options[:src_base] = source_base if source_base.present?

    options
  end

  # These attributes are preserved from the transfer spec provided by shares
  # Options specified in console will be overwritten
  def preserved_shares_attributes
    [
      :cipher,
      :content_protection,
      :content_protection_password,
      :rate_policy,
      :rate_policy_allowed,
      :target_rate_kbps,
      :target_rate_cap_kbps
    ]
  end

  # Submit the transfer spec to the initiating side
  def transfer_request
    spec = JSON.parse(transfer_spec).symbolize_keys

    spec[:tags] = (spec[:tags] || {}).merge(tags_for_job_request || {})

    # Before starting transfer, pull destination out of source/dest pairs and put it in the destination root
    # This is to satisfy Shares
    if shares_transfer?
      # AC - 677 - Destination always to be the file name and Destination root to be folder
      spec[:paths] = spec[:paths].map{ |path| { :source => path['source'],:destination => File.basename(path['source']) } }
    end

    if shares_transfer?
      # Preserve shares transfer attributes
      spec.merge(transfer_request_options).merge(spec.slice(*preserved_shares_attributes))
    else
      spec.merge(transfer_request_options)
    end
  end

  # Transfer request with no upload/download setup
  # Remote side MUST have SSH credentials
  def direct_transfer_request
    request = transfer_request_options
    request = request.merge({
      :destination_root => dest_path,
      :direction => source_initiated? ? 'send' : 'receive',
      :paths => paths,
      :remote_host => remote_login.address,
      :remote_user => remote_login.login,
      :remote_password => remote_login.unencrypted_password,
      :ssh_port => remote_login.port,
      :tags => tags_for_job_request
    })
    # AC-1005 When destination Auth is SSHKey - Pass Unencrypted Private SSH Key and Passphrase
    if remote_login.ssh_key_id.present?
      ssh_key = SshKey.find(remote_login.ssh_key_id)
      request = request.merge({
        :ssh_private_key => ssh_key.unencrypted_private_key,
        :ssh_private_key_passphrase => ssh_key.unencrypted_passphrase
    })
    end
    request
  end
end
