=begin
class hierarchy

NodePollerThread
    ClusterNodePollerActiveSessionUuidsThread
    ClusterNodePollerDataCollectorThread
        ClusterNodePollerFileDataCollectorThread
        ClusterNodePollerSessionDataCollectorThread
    NodePollerActiveSessionUuidsThread
    NodePollerAsyncThread
    NodePollerDataCollectorThread
        NodePollerFileDataCollectorThread
        NodePollerSessionDataCollectorThread
    NodePollerJobSubmitThread
    NodePollerNodeInfoThread
    NodePollerServicesThread
    NodePollerWatchFolderThread
=end

class NodePollerThread
  include PollerLogging

  ALLOWED_INPUT_CLASSES = [NilClass, String, Fixnum, Symbol, TrueClass, FalseClass]
  ALLOWED_INPUT_CLASSES_DESC = ALLOWED_INPUT_CLASSES.collect(&:name).join(', ')

  STATUS_UNINITIALIZED = :uninitialized
  STATUS_RUNNING = :running
  STATUS_COMPLETED = :completed
  STATUS_UNCAUGHT_EXCEPTION = :uncaught_exception
  STATUS_TIMED_OUT = :timed_out
  STATUS_RESULT_AVAILABLE = :result_available


  # ==================================================================
  #
  # Public interface
  #
  # ==================================================================

  def initialize(thread_inputs = {})
    @node_address = thread_inputs[:address]
    reset_thread_vars
    validate_inputs(thread_inputs)
    @inputs = thread_inputs
    @updated_inputs = {}
    log_thread_inputs(@inputs)
  end

  # feed in new information if node changes
  def update_inputs!(new_thread_inputs = {})
    validate_inputs(new_thread_inputs)
    # we don't update @inputs immediately, want to make sure we don't touch it when thread is running
    @updated_inputs = new_thread_inputs
  end

  # maintain the ruby thread
  def process!
    case ruby_thread_status
      when STATUS_UNINITIALIZED
        thread_create!
      when STATUS_RUNNING
        thread_timeout_if_needed!
      when STATUS_COMPLETED
        thread_result_process!
      when STATUS_UNCAUGHT_EXCEPTION
        thread_failure_process!
      when STATUS_RESULT_AVAILABLE
        # nothing
        log_debug 'Result available, awaiting retrieval'
    end
  end

  def result_waiting?
    @result_waiting
  end

  def retrieve_result!
    raise 'no result to retrieve' unless result_waiting?
    return_value = @result.clone rescue @result
    reset_thread_vars
    return_value
  end

  # to assist debugging - perform work_of_thread in same process (don't spawn thread)
  # when using this method, don't call #process!()
  def ad_hoc_execute_unthreaded!
    work_execute!
  end

  def ruby_thread_status
    if @ruby_thread.nil?
      thread_timed_out? ? STATUS_TIMED_OUT : STATUS_UNINITIALIZED
    elsif @ruby_thread.status == false
      result_waiting? ? STATUS_RESULT_AVAILABLE : STATUS_COMPLETED
    elsif @ruby_thread.status.nil?
      STATUS_UNCAUGHT_EXCEPTION
    else
      STATUS_RUNNING
    end
  end

  # ==================================================================
  #
  # Private methods
  #
  # ==================================================================

  private

  # make sure only primitive data types have been passed in, to reduce chance of accidentally invoking ActiveRecord within thread
  def validate_inputs(thread_inputs)
    if thread_inputs.is_a? Hash
      thread_inputs.each_pair do |k, v|
        # check key type
        raise ArgumentError, "#{k.class.name} found. Keys in thread_inputs must be one of the following types: #{ALLOWED_INPUT_CLASSES_DESC}" unless ALLOWED_INPUT_CLASSES.include?(k.class)
        if v.is_a?(Array) || v.is_a?(Hash)
          # recurse into Array or Hash
          validate_inputs(v)
        else
          raise ArgumentError, "#{v.class.name} found. Values in thread_inputs must be one of the following types: #{ALLOWED_INPUT_CLASSES_DESC}" unless ALLOWED_INPUT_CLASSES.include?(v.class)
        end
      end
    elsif thread_inputs.is_a? Array
      thread_inputs.each do |v|
        if v.is_a?(Array) || v.is_a?(Hash)
          # recurse into Array or Hash
          validate_inputs(v)
        else
          raise ArgumentError, "#{v.class.name} found. Values in thread_inputs must be one of the following types: #{ALLOWED_INPUT_CLASSES_DESC}" unless ALLOWED_INPUT_CLASSES.include?(v.class)
        end
      end
    else
      raise ArgumentError, 'thread_inputs must be a hash or array'
    end
  end

  # ==================================================================
  # Override this method in child classes
  # ==================================================================
  #
  # The actual code executed in the interior of the Ruby thread.
  #
  # Within this method:
  #
  #   Don't access database! (no ActiveRecord objects!)
  #   Don't access logger! (use #t_log instead)
  #   Don't call any methods that would access database or logger!
  #
  # Methods of self beginning with 't_' should be safe
  #
  # Method should return data, rely on parent class to handle exceptions
  #
  # In case of error, raise an exception. Note that many errors are
  # already handled by the 't_*' utility methods. (e.g. HTTPUnauthorized)
  #
  # ==================================================================
  def work_of_thread
    raise NotImplementedError, 'method called in abstract base class'
  end


  # wrap work in exception handler, put results into hash
  def work_execute!

    result = work_of_thread
    {
        :success => true,
        :data => result,
        :log => t_log_lines,
        :node_errors_to_clear => t_node_errors_to_clear,
        :node_errors_to_set => t_node_errors_to_set
    }
  rescue
    if $!.class.name =~ /^FaspNodeError::/
      @node_errors_to_set << $!
    end
    {
        :success => false,
        :exception => $!,
        :log => t_log_lines,
        :node_errors_to_clear => t_node_errors_to_clear,
        :node_errors_to_set => t_node_errors_to_set
    }
  end

  # vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv
  # functions for values that are not copied directly from node reply
  # (they require tweaking or come from elsewhere)
  #
  # All should have name in form: spec_*_
  #
  # EXECUTES IN SEPARATE THREAD: AVOID USING DB, LOGGER!
  #

  def spec_file_start_byte_(data)
    if data["start_offset"].to_i == 9223372036854775807
      0
    else
      data["start_offset"].to_i
    end
  end

  def spec_file_bytes_written_(data)
    if data["bytes_written"].to_i == 9223372036854775807
      0
    else
      data["bytes_written"].to_i
    end
  end
  # ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

  # ==================================================================
  #
  # Utility functions for use from within thread
  # (all prefixed with 't_' to serve as a reminder not to call)
  #
  # Within these methods:
  #
  #   Don't access database! (no ActiveRecord objects!)
  #   Don't access logger!
  #   Don't call any methods that would access database or logger!
  #
  # ==================================================================

  # return value if not blank, raise exception if blank
  # keep track of node errors to set or clear after return from thread
  def t_check_for_blank_node_value(value, exception_type)
    if value.blank?
      error = exception_type.new
      t_node_errors_to_set_append error
      raise error
    else
      t_node_errors_to_clear_append exception_type
      value
    end
  end

  # Queries AsperaCentral port, but not using SOAP
  def t_central_driver
    Net::HTTP.new(t_soap_address, t_soap_port)
  end

  def t_central_get(url, accept = '*')
    req = Net::HTTP::Get.new(url)
    req['accept'] = accept
    t_central_request(req)
  end

  def t_central_job_submit(message)
    url = '/services/soap/JobNET-200601/Submit'
    headers = {
        'Content-Type' => 'text/xml;charset=UTF-8',
        'SOAPAction' => '"JobNET-200601#Submit"'
    }
    t_central_post(url, headers, message)
  end

  def t_central_post(url, headers, message)
    req = Net::HTTP::Post.new(url, headers)
    req.body = message
    t_central_request(req)
  end

  def t_central_get_xml(url)
    t_central_get(url, 'application/xml')
  end

  def t_central_request(req)
    t_log "central request: #{req.inspect}"

    begin
      result, body = Timeout::timeout(t_node_timeout_seconds) {
        t_central_driver.request(req)
      }
    rescue Timeout::Error
      raise FaspNodeError::CentralTimeout
    rescue Errno::ETIMEDOUT
      raise FaspNodeError::CentralTimeout
    rescue Errno::ECONNREFUSED
      raise FaspNodeError::CentralConnectionRefused
    rescue NoMethodError
      # catch ruby error when trying to close non-existent socket after
      # connection failure
      if $!.message == %q[undefined method `closed?' for nil:NilClass]
        raise FaspNodeError::CentralUnreachableError
      else
        # reraise exception
        raise
      end
    end

    case result
      when Net::HTTPSuccess then
        t_node_errors_to_clear_append [
                                          FaspNodeError::CentralBadResponseCode,
                                          FaspNodeError::CentralConnectionRefused,
                                          FaspNodeError::CentralUnreachableError,
                                          FaspNodeError::CentralTimeout
                                      ]
        body
      else
        raise FaspNodeError::CentralBadResponseCode, result.code
    end

  end

  def t_decode_json(json)
    t_log "Decoding JSON string #{json}"
    result = ActiveSupport::JSON.decode(json.to_s)
    t_node_errors_to_clear_append(FaspNodeError::NodeApiJsonParseError)
    result
  rescue JSON::ParserError
    raise FaspNodeError::NodeApiJsonParseError, "Error parsing node JSON response: #{$!.message}"
  end

  def t_legacy_node?
    @inputs[:legacy_node]
  end

  # logging function for interior of thread - save to array instead of writing to logger
  def t_log(msg)
    @log_lines << msg if t_log?
  end

  def t_log?
    @inputs[:log]
  end

  def t_log_lines
    @log_lines
  end

  def t_node_address
    @inputs[:address]
  end

  def t_node_api_address
    @inputs[:node_api_address]
  end

  def t_node_api_delete(url, headers, message)
    req = Net::HTTP::Delete.new(url, headers)
    req.body = message
    t_node_api_request(req)
  end

  def t_node_api_driver(use_proxy = true)
    # AC - 827 : To Check for Console Proxy and retrieve Proxy Details
    if use_proxy
      proxy_fields = ConsoleConfig::ConsoleProxy.retrieve_proxy_data("http_proxy", t_node_api_address, ConsoleConfig::ConsoleProxy::PROXY_SOURCE[4])
    end

    if use_proxy && proxy_fields
      http = Net::HTTP.new(t_node_api_address, t_node_api_port, proxy_fields[:host], proxy_fields[:port], proxy_fields[:user], proxy_fields[:pass])
    else
      http = Net::HTTP.new(t_node_api_address, t_node_api_port)
    end

    http.use_ssl = t_node_api_use_ssl?
    if http.use_ssl
      if t_node_api_validate_ssl_cert?
        http.verify_mode = OpenSSL::SSL::VERIFY_PEER
        http.ca_file = File.join(Rails.root, 'config', 'cacert.pem')
      else
        http.verify_mode = OpenSSL::SSL::VERIFY_NONE
      end
    end
    http
  end

  def t_node_api_get_active_cluster_session_ids
    url = '/ops/transfers?view=short_summary&active_only=true&count=100'

    reply = t_node_api_get_json(url)

    # Pull session IDs out of each transfer in the response
    t_decode_json(reply).map{ |transfer| transfer['sessions'].map{ |session| session['id'] } }.flatten
  end

  def t_node_api_get_active_session_ids
    url = '/services/rest/transfers/v1/sessions'

    headers = {'Content-Type' => 'application/json', 'Accept' => 'application/json'}
    message = {
        'session_filter' => {
            :max_result => 1000,
            :session_statuses => ['running', 'paused']
        }
    }.to_json
    reply = t_node_api_post(url, headers, message)

    decoded_reply = t_decode_json(reply)

    if decoded_reply['session_info_result']['result_count'].to_i > 0
      decoded_reply['session_info_result']['session_info'].collect{|s| s['session_uuid']}
    else
      []
    end
  end

  def t_node_api_job_submit(message)
    url = '/services/soap/JobNET-200601/Submit'
    headers = {
        'Content-Type' => 'text/xml;charset=UTF-8',
        'SOAPAction' => '"JobNET-200601#Submit"'
    }
    t_node_api_post(url, headers, message)
  end

  def t_node_api_login
    t_check_for_blank_node_value @inputs[:node_api_login], FaspNodeError::NodeApiLoginMissing
  end

  def t_node_api_password
    t_check_for_blank_node_value @inputs[:node_api_password], FaspNodeError::NodeApiPasswordMissing
  end

  def t_node_api_port
    @inputs[:node_api_port]
  end

  def t_node_api_get(url, accept = '*')
    req = Net::HTTP::Get.new(url)
    req['Accept'] = accept
    req['X-aspera-wf-version'] = '2017_10_23'
    t_node_api_request(req)
  end

  def t_node_api_get_json(url)
    t_node_api_get(url, 'application/json')
  end

  def t_node_api_get_json_and_link_header(url, accept = '*')
    req = Net::HTTP::Get.new(url)
    req['Accept'] = accept
    t_node_api_request(req, true)
  end

  def t_node_api_get_xml(url)
    t_node_api_get(url, 'application/xml')
  end

  def t_node_api_post(url, headers, message, use_proxy = true)
    req = Net::HTTP::Post.new(url, headers)
    req.body = message
    t_node_api_request(req, false, use_proxy)
  end

  def t_node_api_request(req, include_link_header = false, use_proxy = true)
    t_log "node_api_request: #{req.inspect}"
    req.basic_auth t_node_api_login, t_node_api_password

    begin
      result, body = Timeout::timeout(t_node_timeout_seconds) {
        t_node_api_driver(use_proxy).request(req)
      }
    rescue Timeout::Error
      raise FaspNodeError::NodeApiTimeout
    rescue Errno::ETIMEDOUT
      raise FaspNodeError::NodeApiTimeout
    rescue Errno::ECONNREFUSED
      raise FaspNodeError::NodeApiConnectionRefused
    rescue EOFError
      raise FaspNodeError::NodeApiEofError
    rescue NoMethodError
      # catch ruby error when trying to close non-existent socket after
      # connection failure
      if $!.message == %q[undefined method `closed?' for nil:NilClass]
        raise FaspNodeError::NodeApiUnreachableError
      else
        # reraise exception
        raise
      end
    rescue OpenSSL::SSL::SSLError
      if $!.message.include?('certificate verify failed')
        raise FaspNodeError::NodeApiSslCertificateError
      else
        raise FaspNodeError::NodeApiSslError
      end
    end

    case result
      when Net::HTTPSuccess then
        t_node_errors_to_clear_append [
                                          FaspNodeError::NodeApiAuthorizationFailed,
                                          FaspNodeError::NodeApiBadResponseCode,
                                          FaspNodeError::NodeApiConnectionRefused,
                                          FaspNodeError::NodeApiEofError,
                                          FaspNodeError::NodeApiSslCertificateError,
                                          FaspNodeError::NodeApiSslError,
                                          FaspNodeError::NodeApiTimeout,
                                          FaspNodeError::NodeApiUnreachableError
                                      ]
        # Link header is required for clusters because it includes the iteration token
        include_link_header ? {:body => body, :link_header => result['Link']} : body
      when Net::HTTPUnauthorized
        t_node_errors_to_clear_append [
                                          FaspNodeError::NodeApiBadResponseCode,
                                          FaspNodeError::NodeApiConnectionRefused,
                                          FaspNodeError::NodeApiEofError,
                                          FaspNodeError::NodeApiSslCertificateError,
                                          FaspNodeError::NodeApiSslError,
                                          FaspNodeError::NodeApiTimeout,
                                          FaspNodeError::NodeApiUnreachableError
                                      ]
        raise FaspNodeError::NodeApiAuthorizationFailed
      else
        t_node_errors_to_clear_append [
                                          FaspNodeError::NodeApiAuthorizationFailed,
                                          FaspNodeError::NodeApiConnectionRefused,
                                          FaspNodeError::NodeApiEofError,
                                          FaspNodeError::NodeApiSslCertificateError,
                                          FaspNodeError::NodeApiSslError,
                                          FaspNodeError::NodeApiTimeout,
                                          FaspNodeError::NodeApiUnreachableError
                                      ]
        raise FaspNodeError::NodeApiBadResponseCode, result.code
    end
  end

  def t_node_api_use_ssl?
    @inputs[:node_api_use_ssl]
  end

  def t_node_api_validate_ssl_cert?
    @inputs[:node_api_validate_ssl_cert]
  end

  def t_node_errors_to_clear
    @node_errors_to_clear
  end

  def t_node_errors_to_clear_append(types)
    types = [types].flatten
    @node_errors_to_clear |= types
  end

  def t_node_errors_to_set
    @node_errors_to_set
  end

  def t_node_errors_to_set_append(errors)
    @node_errors_to_set
    errors = [errors].flatten
    @node_errors_to_set |= errors
  end

  # timeout for node polling
  def t_node_timeout_seconds
    @inputs[:node_timeout_seconds]
  end

  def t_soap_address
    @inputs[:soap_address]
  end

  def t_soap_port
    @inputs[:soap_port]
  end

  def t_soap_driver
    AsperaSOAP.central_fasp_session_driver(t_soap_address, t_soap_port)
  end

  # timeout for node polling
  def t_thread_timeout_seconds
    @inputs[:thread_timeout_seconds]
  end

  # ==================================================================
  #
  # Thread maintenance
  #
  # ==================================================================

  def reset_thread_vars
    @ruby_thread = nil
    @ruby_thread_started_at = nil
    @result = nil
    @result_waiting = false
    @log_lines = []
    @node_errors_to_clear = []
    @node_errors_to_set = []
  end

  def thread_create!
    unless @updated_inputs.empty?
      log_debug "Updating thread inputs: #{scrubbed_thread_inputs(@updated_inputs).inspect}"
      # only update @inputs when we know thread isn't running
      @inputs = @updated_inputs
      @updated_inputs = nil
    end
    log_debug 'starting thread'
    @ruby_thread_started_at = Time.now
    @ruby_thread = Thread.new do
      work_execute!
    end
  end

  # unusual error, not caught by 'rescue Exception' in thread_create!()
  def thread_failure_process!
    log_error "Unhandled thread failure."
    reset_thread_vars
  end

  def thread_kill!
    log_warn "Killing thread that was started at #{@ruby_thread_started_at}"
    @ruby_thread.kill
    reset_thread_vars
  end

  def thread_result_process!
    errored_step = ''
    thread_result = nil
    return_value = nil
    return_value_available = false

    log_info 'Processing end of thread.'

    errored_step = 'retrieving thread return value'
    thread_result = @ruby_thread.value

    errored_step = 'checking for well-formed result'
    raise ArgumentError, "Malformed result: #{thread_result.class} #{thread_result.inspect}" unless thread_result.is_a?(Hash) && thread_result.has_key?(:success)

    if thread_result[:success]
      errored_step = 'processing thread return value'
      log_info 'Thread returned data'
      log_debug "Thread result: #{thread_result[:data].inspect}"
    else
      errored_step = 'processing thread exception'
      # thread had an exception that was caught
      log_exception thread_result[:exception], 'Thread failed'
    end

    @result = thread_result
    @result_waiting = true
  rescue
    log_exception $!, errored_step
    reset_thread_vars
    return
  end

  def thread_seconds_elapsed
    result = @ruby_thread_started_at && (Time.now - @ruby_thread_started_at)
    result.to_i
  end

  def thread_timed_out?
    result = thread_seconds_elapsed >= t_thread_timeout_seconds
    result
  end

  def thread_timeout_if_needed!
    if thread_timed_out?
      log_warn 'Thread timed out.'
      thread_kill!
    end
  end

  # ==================================================================
  #
  # Misc logging support
  #
  # ==================================================================

  # override in child class if any additional information needs to be redacted (e.g. passwords)
  def log_thread_inputs(input_hash)
    log_debug "new thread, inputs=#{scrubbed_thread_inputs(input_hash).inspect}"
  end

  # extend in child class if any additional information needs to be redacted (e.g. ssh passwords)
  def scrubbed_thread_inputs(input_hash)
    thread_inputs_clone = input_hash.clone
    thread_inputs_clone[:node_api_password]="[FILTERED]"
    thread_inputs_clone
  end

  # ==================================================================
  #
  # Extended UTF-8 handling
  #
  # ==================================================================

  # Encode only UTF-8 characters with 4+ bytes
  def encode_extended_utf8(s)
    return '' if s.nil?

    s.chars.map{ |c| c.bytes.count >= 4 ? encode_utf8(c) : c}.join rescue ''
  end

  # Encode string - eg. encode('asd') => "\\u0061\\u0073\\u0064"
  def encode_utf8(s)
    return '' if s.nil?

    s.unpack('U*').map{ |i| "\\u" + i.to_s(16).rjust(4, '0') }.join
  end

  # Decode string encoded like above
  def decode_utf8(s)
    return '' if s.nil?

    s.gsub!(/\\u(.{4})/) do |match|
      [$1.to_i(16)].pack('U')
    end
  end
end
